package nl.andrewl.aos2_client.render;

import nl.andrewl.aos2_client.Camera;
import nl.andrewl.aos_core.model.Chunk;
import nl.andrewl.aos_core.model.World;
import org.joml.Matrix4f;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

import static org.lwjgl.opengl.GL46.*;

/**
 * The chunk renderer is responsible for managing the shader program, uniforms,
 * and all currently loaded chunk meshes, so that the set of loaded chunks can
 * be rendered each frame.
 */
public class ChunkRenderer {
	private final ChunkMeshGenerator chunkMeshGenerator = new ChunkMeshGenerator();
	private final Queue<Chunk> meshGenerationQueue = new ConcurrentLinkedQueue<>();

	private ShaderProgram shaderProgram;
	private int projectionTransformUniform;
	private int viewTransformUniform;
	private int chunkPositionUniform;

	private final List<ChunkMesh> chunkMeshes = new ArrayList<>();

	public void setupShaderProgram() {
		this.shaderProgram = new ShaderProgram.Builder()
				.withShader("shader/chunk/vertex.glsl", GL_VERTEX_SHADER)
				.withShader("shader/chunk/fragment.glsl", GL_FRAGMENT_SHADER)
				.build();
		shaderProgram.use();
		this.projectionTransformUniform = shaderProgram.getUniform("projectionTransform");
		this.viewTransformUniform = shaderProgram.getUniform("viewTransform");
		this.chunkPositionUniform = shaderProgram.getUniform("chunkPosition");
		int chunkSizeUniform = shaderProgram.getUniform("chunkSize");

		// Set constant uniforms that don't change during runtime.
		glUniform1i(chunkSizeUniform, Chunk.SIZE);
	}

	public void addChunkMesh(Chunk chunk) {
		meshGenerationQueue.add(chunk);
	}

	public void setPerspective(Matrix4f projectionTransform) {
		glUniformMatrix4fv(projectionTransformUniform, false, projectionTransform.get(new float[16]));
	}

	public void draw(Camera cam, World world) {
		while (!meshGenerationQueue.isEmpty()) {
			chunkMeshes.add(new ChunkMesh(meshGenerationQueue.remove(), world, chunkMeshGenerator));
		}
		shaderProgram.use();
		glUniformMatrix4fv(viewTransformUniform, false, cam.getViewTransformData());
		for (var mesh : chunkMeshes) {
			glUniform3iv(chunkPositionUniform, mesh.getPositionData());
			mesh.draw();
		}
	}

	public void free() {
		for (var mesh : chunkMeshes) mesh.free();
		shaderProgram.free();
	}
}
